#include <windows.h>
#include <string.h>

#define ROL(x, n) (((x) << (n)) ^ ((x) >> (64 - (n))))

static unsigned long long state[25], BC[5];

static unsigned long long RoundConstants[24] = {
	0x0000000000000001, 0x0000000000008082, 0x800000000000808a, 0x8000000080008000, 0x000000000000808b, 0x0000000080000001, 0x8000000080008081, 0x8000000000008009,
	0x000000000000008a, 0x0000000000000088, 0x0000000080008009, 0x000000008000000a, 0x000000008000808b, 0x800000000000008b, 0x8000000000008089, 0x8000000000008003,
	0x8000000000008002, 0x8000000000000080, 0x000000000000800a, 0x800000008000000a, 0x8000000080008081, 0x8000000000008080, 0x0000000080000001, 0x8000000080008008
};
static unsigned int RotationConstants[24] = {1, 3, 6, 10, 15, 21, 28, 36, 45, 55, 2, 14, 27, 41, 56, 8, 25, 43, 62, 18, 39, 61, 20, 44};
static unsigned int PiLane[24] = {10, 7, 11, 17, 18, 3, 5, 16, 8, 21, 24, 4, 15, 23, 19, 13, 12, 2, 20, 14, 22, 9, 6, 1};

void KeccakF(unsigned long long *in)
{
	unsigned int x, y;
	unsigned long long temp;
	int laneCount;

	for (x = 0; x < 9; x++)
		state[x] ^= in[x];

	for (laneCount = 0; laneCount < 24; laneCount++) {
		for (x = 0; x < 5; x++)
			BC[x] = state[x] ^ state[x + 5] ^ state[x + 10] ^ state[x + 15] ^ state[x + 20];

		for (x = 0; x < 5; x++) {
			temp = BC[(x + 4) % 5] ^ ROL(BC[(x + 1) % 5], 1);
			for (y = 0; y < 25; y += 5)
				state[x + y] ^= temp;
		}

		temp = state[1];
		for (x = 0; x < 24; x++) {
			BC[0] = state[PiLane[x]];
			state[PiLane[x]] = ROL(temp, RotationConstants[x]);
			temp = BC[0];
		}

		for (y = 0; y < 25; y += 5) {
			for (x = 0; x < 5; x++)
				BC[x] = state[y + x];
			for (x = 0; x < 5; x++)
				state[y + x] = BC[x] ^ (~BC[(x + 1) % 5] & BC[(x + 2) % 5]);
		}

		state[0] ^= RoundConstants[laneCount];
	}
}

void sha3(unsigned char *out, unsigned char *in, unsigned long inlen, int init, int finalize)
{
	unsigned char temp[72];

	if (init)
		memset(state, 0, sizeof state);

	while (inlen >= 72) {
		KeccakF((unsigned long long*)in);
		inlen -= 72;
		in += 72;
	}

	if (finalize) {
		memcpy(temp, in, inlen);
		temp[inlen++] = 1;
		memset(temp + inlen, 0, 72 - inlen);
		temp[71] |= 0x80;
		KeccakF((unsigned long long*)temp);
		memcpy(out, state, 64);
		SecureZeroMemory(temp, sizeof temp);
		SecureZeroMemory(state, sizeof state);
		SecureZeroMemory(BC, sizeof BC);
	}
}
